########
# Copyright (c) 2014-2022 Cloudify Platform Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import yaml

from .. import LintProblem

from ..generators import CfyNode
from ..utils import process_relevant_tokens

VALUES = []

ID = 'dsl_definitions'
TYPE = 'token'
CONF = {'allowed-values': list(VALUES), 'check-keys': bool}
DEFAULT = {'allowed-values': ['true', 'false'], 'check-keys': True}


@process_relevant_tokens(CfyNode, ['dsl_definitions'])
def check(token=None, **_):
    for dsl_definition in token.node.value:
        if not isinstance(dsl_definition[0].value,
                          str) or dsl_definition[0].value.isdigit():
            yield LintProblem(
                token.line,
                None,
                'dsl definition should be a string and '
                'should not start with a numeric character: {}'
                .format(dsl_definition[0].value)
            )
        if not isinstance(dsl_definition[1], yaml.nodes.MappingNode):
            yield LintProblem(
                token.line,
                None,
                'dsl definition {} content must be a dict: {}'
                .format(dsl_definition[0].value,
                        type(dsl_definition[1].value))
            )



########
# Copyright (c) 2014-2022 Cloudify Platform Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .. import LintProblem

from ..generators import CfyNode
from ..utils import process_relevant_tokens, context

VALUES = []

ID = 'dsl_version'
TYPE = 'token'
CONF = {'allowed-values': list(VALUES), 'check-keys': bool}
DEFAULT = {'allowed-values': ['true', 'false'], 'check-keys': True}

LINTING_VERSIONS = ['cloudify_dsl_1_3', 'cloudify_dsl_1_4']
INVALID_3_1 = [
    'blueprint_id',
    'deployment_id',
    'capability_value',
    'scaling_group',
    'secret_key',
    'node_id',
    'node_type',
    'node_instance'
]


@process_relevant_tokens(CfyNode, ['tosca_definitions_version', 'type'])
def check(token=None, **_):
    if token.prev and token.prev.node.value == 'tosca_definitions_version':
        context['dsl_version'] = token.node.value
        yield from validate_supported_dsl_version(token.node.value, token.line)
    if token.prev and token.prev.node.value == 'type':
        yield from validate_dsl_version_31(
            token.node.value, token.line, context.get('dsl_version'))


def validate_supported_dsl_version(value, line):
    if value not in LINTING_VERSIONS:
        yield LintProblem(
            line,
            None,
            "dsl_version not supported: {} ".format(value)
        )


def validate_dsl_version_31(value, line, dsl_version):
    if dsl_version and value in INVALID_3_1:
        yield LintProblem(
            line,
            None,
            "invalid type for {}: {} ".format(dsl_version, value)
        )



########
# Copyright (c) 2014-2022 Cloudify Platform Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import io
import os

from ..logger import logger
from ..yamllint_ext.config import YamlLintConfigExt
from ..yamllint_ext import (run, rules)

from .. import cli


@cli.command()
@cli.options.blueprint_path
@cli.options.config
def lint(blueprint_path, config):
    yaml_config = YamlLintConfigExt(content=config, yamllint_rules=rules)

    report = create_report_for_file(blueprint_path, yaml_config)
    cnt = 0
    for item in report:
        message = '{0: <4}: {1:>4}'.format(item.line, item.message)
        if cnt == 0:
            logger.info('The following linting errors were found: ')
            cnt += 1
        if item.level == 'warning':
            logger.warning(message)
        elif item.level == 'error':
            logger.error(message)
        else:
            logger.info(message)


def create_report_for_file(file_path, conf):
    if not os.path.exists(file_path):
        raise RuntimeError('File path does not exist: {}.'.format(file_path))
    logger.info('Linting blueprint: {}'.format(file_path))
    with io.open(file_path, newline='') as f:
        return run(f, conf)



########
# Copyright (c) 2014-2022 Cloudify Platform Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import click

CLICK_CONTEXT_SETTINGS = dict(
    help_option_names=['-h', '--help'])


def init():
    pass


def group(name):
    return click.group(name=name, context_settings=CLICK_CONTEXT_SETTINGS)


def command(*args, **kwargs):
    return click.command(*args, **kwargs)


class Options(object):
    def __init__(self):
        self.blueprint_path = click.option(
            '-b',
            '--blueprint-path',
            default='blueprint.yaml',
            type=click.Path(),
            multiple=False,
            show_default='blueprint.yaml',
            help='Path to the blueprint file that you want to lint.')

        self.config = click.option(
            '-c',
            '--config',
            default=None,
            type=click.Path(),
            multiple=False,
            help='ability to use configuration file or options.')


options = Options()



########
# Copyright (c) 2014-2022 Cloudify Platform Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import yaml


def spaces_after(token, prev, next, min=-1, max=-1,
                 min_desc=None, max_desc=None):
    if next is not None and token.end_mark.line == next.start_mark.line:
        spaces = next.start_mark.pointer - token.end_mark.pointer
        if max != - 1 and spaces > max:
            return LintProblem(token.start_mark.line + 1,
                               next.start_mark.column, max_desc)
        elif min != - 1 and spaces < min:
            return LintProblem(token.start_mark.line + 1,
                               next.start_mark.column + 1, min_desc)


def spaces_before(token, prev, next, min=-1, max=-1,
                  min_desc=None, max_desc=None):
    if (prev is not None and prev.end_mark.line == token.start_mark.line and
            # Discard tokens (only scalars?) that end at the start of next line
            (prev.end_mark.pointer == 0 or
             prev.end_mark.buffer[prev.end_mark.pointer - 1] != '\n')):
        spaces = token.start_mark.pointer - prev.end_mark.pointer
        if max != - 1 and spaces > max:
            return LintProblem(token.start_mark.line + 1,
                               token.start_mark.column, max_desc)
        elif min != - 1 and spaces < min:
            return LintProblem(token.start_mark.line + 1,
                               token.start_mark.column + 1, min_desc)


class LintProblem(object):
    """Represents a linting problem found by yamllint."""
    def __init__(self, line, column, desc='<no description>', rule=None):
        #: Line on which the problem was found (starting at 1)
        self.line = line
        #: Column on which the problem was found (starting at 1)
        self.column = column or 0
        #: Human-readable description of the problem
        self.desc = desc
        #: Identifier of the rule that detected the problem
        self.rule = rule
        self.level = None

    @property
    def message(self):
        if self.rule is not None:
            return '({}): {}'.format(self.rule, self.desc)
        return self.desc

    def __eq__(self, other):
        return (self.line == other.line and
                self.column == other.column and
                self.rule == other.rule)

    def __lt__(self, other):
        return (self.line < other.line or
                (self.line == other.line and self.column < other.column))

    def __repr__(self):
        return '%d:%d: %s' % (self.line, self.column, self.message)


def get_syntax_error(buffer):
    try:
        list(yaml.parse(buffer, Loader=yaml.BaseLoader))
    except yaml.error.MarkedYAMLError as e:
        problem = LintProblem(e.problem_mark.line + 1,
                              e.problem_mark.column + 1,
                              'syntax error: ' + e.problem + ' (syntax)')
        problem.level = 'error'
        return problem



########
# Copyright (c) 2014-2022 Cloudify Platform Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import re
import yaml
from packaging import version
from urllib.parse import urlparse

from .. import LintProblem

from ..generators import CfyNode
from ..utils import process_relevant_tokens

VALUES = []

ID = 'imports'
TYPE = 'token'
CONF = {'allowed-values': list(VALUES), 'check-keys': bool}
DEFAULT = {'allowed-values': ['true', 'false'], 'check-keys': True}


@process_relevant_tokens(CfyNode, 'imports')
def check(token=None, **_):
    for import_item in token.node.value:
        yield from validate_string(import_item, token.line)
        yield from validate_import_items(import_item, token.line)


def validate_import_items(item, line):

    url = urlparse(item.value)

    if url.scheme not in ['http', 'https', 'plugin']:
        if not url.scheme and url.path.split('/')[-1].endswith('.yaml'):
            if not os.path.exists(url.path) and \
                    url.path != 'cloudify/types/types.yaml':
                yield LintProblem(
                    line,
                    None,
                    'relative import '
                    'declared but the file path does not exist.'
                )
        else:
            yield LintProblem(
                line,
                None,
                'invalid import. {} scheme not accepted'.format(url.scheme)
            )
    if url.scheme in ['plugin'] and url.path in ['cloudify-openstack-plugin']:
        yield from check_openstack_plugin_version(url, line)
    elif url.scheme in ['https', 'https'] and not url.path.endswith('.yaml'):
        yield LintProblem(
            line,
            None,
            'invalid import. {}'.format(url)
        )


def validate_string(item, line):
    if not isinstance(item, yaml.nodes.ScalarNode):
        yield LintProblem(line, None, 'import is not a string.')


def check_openstack_plugin_version(url, line):
    version_openstack = url.query.split(',')
    if version_openstack[0]:
        for str_version in version_openstack:
            only_version = re.findall('(\\d+.\\d+.\\d+)', str_version)
            if version.parse(only_version[0]) >= version.parse('3.0.0') and \
                    "<=" not in str_version:
                return
        yield LintProblem(
            line, None,
            'Cloudify Openstack Plugin version {} is deprecated.'
            ' Please update to Openstack version 3 or higher. '
            'Below are suggested node type changes.'
            ' For more information on conversion to Openstack Plugin v3, '
            'Please click on this link - https://docs.cloudify.co/latest/'
            'working_with/official_plugins/infrastructure/openstackv3/'
            .format(version_openstack))



########
# Copyright (c) 2014-2022 Cloudify Platform Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

BLUEPRINT_MODEL = {
    'tosca_definitions_version': None,
    'imports': {},
    'inputs': {},
    'dsl_definitions': {},
    'node_templates': {},
    'capabilities': {},
}

NODE_TEMPLATE_MODEL = {
    'type': '',
    'properties': {},
    'interfaces': {},
    'relationships': {},
    'capabilities': {},
}

LATEST_PLUGIN_YAMLS = {
    'cloudify-aws-plugin': 'https://github.com/cloudify-cosmo/cloudify-aws-plugin/releases/download/latest/plugin.yaml', # noqa
    'cloudify-azure-plugin': 'https://github.com/cloudify-cosmo/cloudify-azure-plugin/releases/download/latest/plugin.yaml', # noqa
    'cloudify-starlingx-plugin': 'https://github.com/cloudify-cosmo/cloudify-starlingx-plugin/releases/download/latest/plugin.yaml', # noqa
    'cloudify-gcp-plugin': 'https://github.com/cloudify-cosmo/cloudify-gcp-plugin/releases/download/latest/plugin.yaml', # noqa
    'cloudify-openstack-plugin': 'https://github.com/cloudify-cosmo/cloudify-openstack-plugin/releases/download/latest/plugin.yaml', # noqa
    'cloudify-vsphere-plugin': 'https://github.com/cloudify-cosmo/cloudify-vsphere-plugin/releases/download/latest/plugin.yaml', # noqa
    'cloudify-terraform-plugin': 'https://github.com/cloudify-cosmo/cloudify-terraform-plugin/releases/download/latest/plugin.yaml', # noqa
    'cloudify-terragrunt-plugin': 'https://github.com/cloudify-cosmo/cloudify-terragrunt-plugin/releases/download/latest/plugin.yaml', # noqa
    'cloudify-ansible-plugin': 'https://github.com/cloudify-cosmo/cloudify-ansible-plugin/releases/download/latest/plugin.yaml', # noqa
    'cloudify-kubernetes-plugin': 'https://github.com/cloudify-cosmo/cloudify-kubernetes-plugin/releases/download/latest/plugin.yaml', # noqa
    'cloudify-docker-plugin': 'https://github.com/cloudify-cosmo/cloudify-docker-plugin/releases/download/latest/plugin.yaml', # noqa
    'cloudify-netconf-plugin': 'https://github.com/cloudify-cosmo/cloudify-netconf-plugin/releases/download/latest/plugin.yaml', # noqa
    'cloudify-fabric-plugin': 'https://github.com/cloudify-cosmo/cloudify-fabric-plugin/releases/download/latest/plugin.yaml', # noqa
    'cloudify-libvirt-plugin': 'https://github.com/cloudify-incubator/cloudify-libvirt-plugin/releases/download/latest/plugin.yaml', # noqa
    'cloudify-utilities-plugin': 'https://github.com/cloudify-incubator/cloudify-utilities-plugin/releases/download/latest/plugin.yaml', # noqa
    'cloudify-host-pool-plugin': 'https://github.com/cloudify-cosmo/cloudify-host-pool-plugin/releases/download/latest/plugin.yaml', # noqa
    'cloudify-vcloud-plugin': 'https://github.com/cloudify-cosmo/cloudify-vcloud-plugin/releases/download/latest/plugin.yaml', # noqa
    'cloudify-helm-plugin': 'https://github.com/cloudify-incubator/cloudify-helm-plugin/releases/download/latest/plugin.yaml' # noqa
}

DEFAULT_NODE_TYPES = [
    'cloudify.nodes.Port',
    'cloudify.nodes.Root',
    'cloudify.nodes.Tier',
    'cloudify.nodes.Router',
    'cloudify.nodes.Subnet',
    'cloudify.nodes.Volume',
    'cloudify.nodes.Network',
    'cloudify.nodes.Compute',
    'cloudify.nodes.Container',
    'cloudify.nodes.VirtualIP',
    'cloudify.nodes.FileSystem',
    'cloudify.nodes.ObjectStorage',
    'cloudify.nodes.LoadBalancer',
    'cloudify.nodes.SecurityGroup',
    'cloudify.nodes.SoftwareComponent',
    'cloudify.nodes.DBMS',
    'cloudify.nodes.Database',
    'cloudify.nodes.WebServer',
    'cloudify.nodes.ApplicationServer',
    'cloudify.nodes.MessageBusServer',
    'cloudify.nodes.ApplicationModule',
    'cloudify.nodes.CloudifyManager',
    'cloudify.nodes.Component',
    'cloudify.nodes.ServiceComponent',
    'cloudify.nodes.SharedResource',
    'cloudify.nodes.Blueprint',
    'cloudify.nodes.PasswordSecret'
]

DEFAULT_RELATIONSHIPS = [
    'cloudify.relationships.depends_on',
    'cloudify.relationships.connected_to',
    'cloudify.relationships.contained_in',
    'cloudify.relationships.depends_on_lifecycle_operation',
    'cloudify.relationships.depends_on_shared_resource',
    'cloudify.relationships.connected_to_shared_resource',
    'cloudify.relationships.file_system_depends_on_volume',
    'cloudify.relationships.file_system_contained_in_compute'
]

DEFAULT_TYPES = {
    'node_types': DEFAULT_NODE_TYPES,
    'relationships': DEFAULT_RELATIONSHIPS,
}



########
# Copyright (c) 2014-2022 Cloudify Platform Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .. import LintProblem

from ..generators import CfyNode
from ..utils import process_relevant_tokens

VALUES = []

ID = 'node_types'
TYPE = 'token'
CONF = {'allowed-values': list(VALUES), 'check-keys': bool}
DEFAULT = {'allowed-values': ['true', 'false'], 'check-keys': True}


@process_relevant_tokens(CfyNode, 'node_types')
def check(token=None, **_):
    for node_type in token.node.value:
        yield from node_type_follows_naming_conventions(
            node_type[0].value, token.line)


def node_type_follows_naming_conventions(value, line):
    split_node_type = value.split('.')
    last_key = split_node_type.pop()
    if not {'cloudify', 'nodes'} <= set(split_node_type):
        yield LintProblem(
            line,
            None,
            "node types should following naming convention cloudify.nodes.*: "
            "{}".format(value))
    if not good_camel_case(last_key, split_node_type):
        new_value = '.'.join(
            [k.lower() for k in split_node_type]) + '.{}'.format(last_key)
        yield LintProblem(
            line,
            None,
            "incorrect camel case {}. Suggested: {} ".format(value, new_value))


def good_camel_case(last_key, split_node_type):
    if not last_key[0].isupper():
        return False
    for key in split_node_type:
        if key[0].isupper():
            return False
    return True



########
# Copyright (c) 2014-2022 Cloudify Platform Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import yaml

from .. import LintProblem
from ..generators import CfyNode
from ..utils import (
    INTRINSIC_FNS,
    recurse_mapping,
    context as ctx, process_relevant_tokens)

VALUES = []

ID = 'inputs'
TYPE = 'token'
CONF = {'allowed-values': list(VALUES), 'check-keys': bool}
DEFAULT = {'allowed-values': ['true', 'false'], 'check-keys': True}


@process_relevant_tokens(CfyNode, ['inputs', 'get_input'])
def check(token=None, **_):
    if token.prev.node.value == 'inputs':
        for item in token.node.value:
            input_obj = CfyInput(item)
            if input_obj.not_input():
                continue
            ctx['inputs'].update(input_obj.__dict__())
            yield from validate_inputs(input_obj, input_obj.line or token.line)

    if token.prev.node.value == 'get_input':
        if isinstance(token.node.value, list):
            if isinstance(token.node.value[0], yaml.nodes.ScalarNode):
                if token.node.value[0].value not in ctx['inputs']:
                    yield LintProblem(
                        token.line,
                        None,
                        'undefined input {}'
                        .format(token.node.value[0].value))
            if isinstance(token.node.value[0], tuple):
                if token.node.value[0][0] not in ctx['inputs']:
                    yield LintProblem(
                        token.line,
                        None,
                        'undefined input "{}"'.format(token.node.value[0][0]))
        else:
            if token.node.value not in ctx['inputs']:
                yield LintProblem(
                    token.line,
                    None,
                    'undefined input "{}"'.format(token.node.value))


def validate_inputs(input_obj, line):
    if not input_obj.input_type:
        message = 'input "{}" does not specify a type. '.format(input_obj.name)
        if input_obj.default:
            if isinstance(input_obj.default, dict):
                for key in input_obj.default.keys():
                    if key in INTRINSIC_FNS:
                        input_obj.default = None
                if isinstance(input_obj.default, dict):
                    message += 'The correct type could be "dict".'
            if isinstance(input_obj.default, str):
                message += 'The correct type could be "string".'
            if isinstance(input_obj.default, bool):
                message += 'The correct type could be "boolean".'
            if isinstance(input_obj.default, list):
                message += 'The correct type could be "list".'
        yield LintProblem(line, None, message)


class CfyInput(object):
    def __init__(self, nodes):
        self._line = None
        self.name, self.mapping = self.get_input(nodes)
        for key in list(self.mapping.keys()):
            if key not in ['type', 'default', 'description', 'constraints']:
                del self.mapping[key]
        self.input_type = self.mapping.get('type')
        self.description = self.mapping.get('description')
        self._default = self.mapping.get('default')
        self.constraints = self.mapping.get('constraints')

    @property
    def default(self):
        return self._default

    @default.setter
    def default(self, value):
        self._default = value

    @property
    def line(self):
        return self._line

    def not_input(self):
        return all([not k for k in self.mapping.values()])

    def __dict__(self):
        return {
            self.name: self.mapping
        }

    def get_input(self, nodes):
        if len(nodes) != 2:
            name = None
            mapping = None
        else:
            name = self.get_input_name(nodes[0])
            mapping = self.get_input_mapping(nodes[1])
        return name, mapping

    def get_input_name(self, node):
        if isinstance(node, yaml.nodes.ScalarNode):
            self._line = node.end_mark.line + 1
            return node.value

    def get_input_mapping(self, node):
        mapping = {
            'type': None,
            'default': None,
            'description': None,
            'constraints': None,
        }
        if isinstance(node, yaml.nodes.MappingNode):
            for tup in node.value:
                if not len(tup) == 2:
                    continue
                mapping_name = tup[0].value
                mapping_value = self.get_mapping_value(
                    mapping_name, tup[1].value)
                mapping[mapping_name] = mapping_value
        return mapping

    def get_mapping_value(self, name, value):
        if name not in ['default', 'constraints']:
            return value
        else:
            return recurse_mapping(value)



########
# Copyright (c) 2014-2022 Cloudify Platform Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import io
import os
import re
import json
import yaml
import urllib.request
from urllib.parse import urlparse
from packaging.version import parse as version_parse

from .cloudify.models import NodeTemplate
from .constants import (
    BLUEPRINT_MODEL,
    DEFAULT_TYPES,
    LATEST_PLUGIN_YAMLS,
    NODE_TEMPLATE_MODEL)

INTRINSIC_FNS = [
    'merge',
    'concat',
    'get_sys',
    'get_input',
    'get_label',
    'get_secret',
    'string_find',
    'string_split',
    'string_lower',
    'string_upper',
    'get_property',
    'get_attribute',
    'string_replace',
    'get_capability',
    'get_environment_capability',
]

context = {
    'imports': [],
    'dsl_version': None,
    'inputs': {},
    'node_templates': {},
    'node_types': {},
    'capabilities': {},
    'outputs': {},
    'current_tokens_line': 0
}


def assign_current_top_level(elem):
    if isinstance(elem.curr, yaml.tokens.ScalarToken) and \
            elem.curr.value in BLUEPRINT_MODEL and \
            isinstance(elem.nextnext,
                       yaml.tokens.BlockMappingStartToken):
        return elem.curr.value
    elif isinstance(elem.curr, yaml.tokens.BlockEndToken) and \
            isinstance(elem.nextnext, yaml.tokens.ScalarToken) and \
            elem.nextnext.value in BLUEPRINT_MODEL:
        return ''


def assign_nested_node_template_level(elem):
    if not isinstance(elem.curr, yaml.tokens.ScalarToken):
        return
    if elem.curr.value not in NODE_TEMPLATE_MODEL:
        return
    if isinstance(elem.nextnext, (yaml.tokens.BlockMappingStartToken,
                                  yaml.tokens.BlockEntryToken)):
        return elem.curr.value


def update_model(_elem):
    """Tracking a Cloudify Model inside YAMLLINT context.

    :param _elem:
    :return:
    """
    # print(vars(_elem))
    context['current_tokens_line'] = _elem.line_no
    if stop_document(_elem):
        # The document is finished.
        return
    # We are in the middle of the document.
    top_level = assign_current_top_level(_elem)
    node_template(_elem)
    if skip_inputs_in_node_templates(_elem):
        return
    elif isinstance(top_level, str):
        context['current_top_level'] = top_level  # noqa


def stop_document(_elem):
    if isinstance(_elem.curr, yaml.tokens.StreamStartToken):
        # This is the start of the YAML document.
        context['model'] = BLUEPRINT_MODEL
        context['current_top_level'] = None  # noqa
    elif isinstance(_elem.curr, yaml.tokens.StreamEndToken):
        # This is the end of the YAML document.
        del context['model']
        return True
    return False


def node_template(_elem):
    if context.get('current_top_level') == 'node_templates':
        # When we are looking at Node Templates, we may
        nt = assign_nested_node_template_level(_elem)
        if isinstance(nt, str):
            context['node_template_level'] = nt
    else:
        context['node_template_level'] = None


def skip_inputs_in_node_templates(top_level):
    return context.get('current_top_level') == 'node_templates' and \
           top_level == 'inputs'


def get_json_from_marketplace(url):
    try:
        resp = urllib.request.urlopen(url)
    except urllib.error.HTTPError:
        return {}
    body = resp.read()
    return json.loads(body)


def get_plugin_id_from_marketplace(plugin_name):
    url_plugin_id = 'https://marketplace.cloudify.co/plugins?name={}'.format(
        plugin_name)
    json_resp = get_json_from_marketplace(url_plugin_id)
    if 'items' in json_resp:
        if len(json_resp['items']) == 1:
            return json_resp['items'][0]['id']


def get_plugin_versions_from_marketplace(plugin_id):
    url_plugin_version = 'https://marketplace.cloudify.co/' \
                         'plugins/{}/versions?'.format(plugin_id)
    json_resp = get_json_from_marketplace(url_plugin_version)
    if 'items' in json_resp:
        versions = [item['version'] for item in json_resp['items']]
        return sorted(versions, key=lambda x: version_parse(x))
    return []


def get_plugin_release_spec_from_marketplace(plugin_id, plugin_version):
    release_url = 'https://marketplace.cloudify.co/plugins/{}/{}'.format(
        plugin_id, plugin_version)
    return get_json_from_marketplace(release_url)


def validate_versions(versions, validations):
    for neq in validations['!=']:
        if neq in versions:
            versions.remove(neq)
    for gteq in validations['>=']:
        parsed_gteq = version_parse(gteq)
        versions = [v for v in versions if version_parse(v) >= parsed_gteq]
    for lteq in validations['<=']:
        parsed_lteq = version_parse(lteq)
        versions = [v for v in versions if version_parse(v) <= parsed_lteq]
    for gt in validations['>']:
        parsed_gt = version_parse(gt)
        versions = [v for v in versions if version_parse(v) > parsed_gt]
    for lt in validations['<=']:
        parsed_lt = version_parse(lt)
        versions = [v for v in versions if version_parse(v) < parsed_lt]
    return versions


def get_validations(version_constraints):
    validations = {
        '==': [],
        '!=': [],
        '>=': [],
        '<=': [],
        '>': [],
        '<': [],
    }
    # Organize the version constraints so we get a dict like this:
    # {
    #    '==': [],
    #    '!=': ['1.0'],
    #    '>=': ['0.8', 0.9'],
    #    '<=': ['1.1'],
    # }
    for version_constraint in version_constraints:
        for validation in validations.keys():
            if version_constraint.startswith(validation):
                validated_version_constraint = version_constraint.split(
                    validation)
                if len(validated_version_constraint) == 2:
                    validations[validation].append(
                        validated_version_constraint[1])
                    continue
    return validations


def get_version_constraints(plugin_name, plugin_version_string):
    version_constraints = list(
        # Get rid of irrelevant stringy stuff.
        filter(
            lambda item: item, re.split(
                'plugin:| |{}|,'.format(plugin_name),
                plugin_version_string)
        )
    )
    # re.split is afraid of this one.
    if '?version' in version_constraints:
        version_constraints.remove('?version')
    if 'version=' in version_constraints:
        version_constraints.remove('version=')
    return version_constraints


def get_plugin_spec(plugin_version_string, plugin_name):
    version_constraints = get_version_constraints(
        plugin_name, plugin_version_string)
    validations = get_validations(version_constraints)

    plugin_id = get_plugin_id_from_marketplace(plugin_name)
    versions = get_plugin_versions_from_marketplace(plugin_id)

    if len(validations['==']) == 1 and validations['=='][0] in versions:
        return get_plugin_release_spec_from_marketplace(
            plugin_id, validations['=='][0])
    versions = validate_versions(versions, validations)
    if len(versions):
        return get_plugin_release_spec_from_marketplace(
            plugin_id, versions[-1])


def get_plugin_yaml_url(plugin_import):
    plugin_name, plugin_spec = _get_plugin_spec(plugin_import)
    if not plugin_spec:
        return LATEST_PLUGIN_YAMLS.get(plugin_name)
    elif len(plugin_spec.get('yaml_urls', [])):
        return plugin_spec['yaml_urls'][0]['url']


def _get_plugin_spec(plugin_import):
    parsed_import_item = urlparse(plugin_import)
    plugin_name = parsed_import_item.path
    return plugin_name, get_plugin_spec(parsed_import_item.query, plugin_name)


def get_node_types_for_plugin_import(plugin_import):
    plugin_name, plugin_spec = _get_plugin_spec(plugin_import)
    plugin_version = None
    if plugin_spec:
        plugin_version = plugin_spec['version']
    return get_node_types_for_plugin_version(plugin_name, plugin_version)


def get_node_types_for_plugin_version(plugin_name, plugin_version):
    url = 'https://marketplace.cloudify.co/node-types?' \
          '&plugin_name={}' \
          '&plugin_version={}'.format(plugin_name, plugin_version)
    result = get_json_from_marketplace(url)
    node_types = {}
    for item in result['items']:
        node_types[item['type']] = item
    return node_types


def import_cloudify_yaml(import_item, base_path=None):
    result = {}
    parsed_import_item = urlparse(import_item)
    if parsed_import_item.scheme == 'plugin':
        result['node_types'] = get_node_types_for_plugin_import(import_item)
    if parsed_import_item.scheme in ['http', 'https']:
        page = urllib.request.Request(import_item,
                                      headers={'User-Agent': 'Mozilla/5.0'})
        infile = urllib.request.urlopen(page).read()
        result = yaml.safe_load(infile)
    elif import_item == 'cloudify/types/types.yaml':
        result = DEFAULT_TYPES
    elif base_path and os.path.exists(os.path.join(base_path, import_item)):

        with open(os.path.join(base_path, import_item), 'r') as stream:
            result = yaml.safe_load(stream)
    elif os.path.exists(import_item):
        with open(import_item, 'r') as stream:
            result = yaml.safe_load(stream)
    result = result or {}
    for k in result.keys():
        left = 'imported_{}'.format(k)
        if left not in context:
            context[left] = result[k]
        elif isinstance(context[left], list):
            context[left].extend(result[k])
        elif isinstance(context[left], str):
            if context[left] != result[k]:
                raise \
                    Exception('There is no match between {context} '
                              'and {result} '
                              .format(context=context[left], result=result[k]))
        else:
            context[left].update(result[k])


def setup_types(buffer=None, data=None, base_path=None):
    data = data or yaml.safe_load(buffer)
    for imported in data.get('imports', {}):
        import_cloudify_yaml(imported, base_path=base_path)


def setup_node_templates(elem):
    if 'node_templates' not in context:
        context['node_templates'] = {}
    if elem.prev and elem.prev.node.value == 'node_templates':
        for item in elem.node.value:
            node_template = setup_node_template(item)
            if node_template.name not in context:
                context['node_templates'].update({
                    node_template.name: node_template
                })
    elem.node_templates = context['node_templates']


def setup_node_template(list_item):
    if len(list_item) == 2:
        if isinstance(list_item[0], yaml.nodes.ScalarNode) and \
                isinstance(list_item[1], yaml.nodes.MappingNode):
            node_template = NodeTemplate(list_item[0].value)
            node_template.node_type = setup_node_type(list_item[1].value)
            return node_template


def setup_node_type(value):
    return value[0][1].value


def mapping_is_two_length_intrinsic_function(mapping):
    if len(mapping) == 2 and not isinstance(mapping[0], tuple):
        try:
            if mapping[0].value in INTRINSIC_FNS:
                return True
        except AttributeError:
            return False


def mapping_is_one_length_intrinsic_function_tuple(mapping):
    if len(mapping) == 1 and isinstance(mapping[0], tuple):
        if len(mapping[0]) == 2 and mapping[0][0].value in INTRINSIC_FNS:
            return True


def mapping_is_one_length_intrisic_function_mapping_node(mapping):
    if len(mapping) == 1 and isinstance(mapping[0],
                                        yaml.nodes.MappingNode):
        try:
            if len(mapping[0].value) == 2 and \
                   mapping[0].value[0].value in INTRINSIC_FNS:
                return True
        except AttributeError:
            return False


def recurse_mapping(mapping):
    if isinstance(mapping, dict):
        new_dict = {}
        for k, v in mapping.items():
            new_dict[k] = recurse_mapping(v)
        return new_dict
    elif isinstance(mapping, (list, tuple)):
        new_list = []
        if mapping_is_two_length_intrinsic_function(mapping):
            return recurse_mapping({mapping[0].value: mapping[1].value})
        if mapping_is_one_length_intrinsic_function_tuple(mapping):
            return recurse_mapping(
                {
                    mapping[0][0].value: mapping[0][1].value
                }
            )
        if mapping_is_one_length_intrisic_function_mapping_node(mapping):
            return recurse_mapping(
                {
                    mapping[0].value[0].value: mapping[0].value[1].value
                }
            )
        for item in mapping:
            new_list.append(recurse_mapping(item))
        return new_list
    elif not isinstance(mapping, yaml.nodes.Node):
        return mapping
    elif isinstance(mapping, yaml.nodes.ScalarNode):
        return mapping.value
    elif isinstance(mapping, yaml.nodes.SequenceNode):
        new_list = []
        for item in mapping.value:
            new_list.append(recurse_mapping(item))
        return new_list
    elif isinstance(mapping, yaml.nodes.MappingNode):
        new_dict = {}
        new_list = []
        for item in mapping.value:
            if isinstance(item, (list, tuple)) and len(item) == 2:
                key = item[0].value
                value = recurse_mapping(item[1].value)
                new_dict[key] = value
            else:
                new_list.append(item)
        if new_dict:
            return new_dict
        return new_list


def process_relevant_tokens(model, keyword):
    def wrapper_outer(function):
        def wrapper_inner(*args, **kwargs):
            token = kwargs.get('token')
            if isinstance(token, model):
                if isinstance(keyword, str):
                    if token.prev and token.prev.node.value == keyword:
                        yield from function(*args, **kwargs)
                if isinstance(keyword, list):
                    if token.prev and token.prev.node.value in keyword:
                        yield from function(*args, **kwargs)
        return wrapper_inner
    return wrapper_outer


def update_dict_values_recursive(default_dict, name_file_config):
    with io.open(name_file_config):
        f = open("config.yaml", "r")
        user_dict = f.read()

    default_dict = yaml.load(default_dict)
    user_dict = yaml.load(user_dict)

    if user_dict and default_dict:
        for key, value in user_dict.items():
            if value is dict:
                update_dict_values_recursive(default_dict[key], value)
            if value:
                default_dict[key] = value
    return default_dict



########
# Copyright (c) 2014-2022 Cloudify Platform Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import yaml

from .. import LintProblem
from ..generators import CfyNode
from ..utils import process_relevant_tokens, recurse_mapping, context as ctx

VALUES = []

ID = 'capabilities'
TYPE = 'token'
CONF = {'allowed-values': list(VALUES), 'check-keys': bool}
DEFAULT = {'allowed-values': ['true', 'false'], 'check-keys': True}


@process_relevant_tokens(CfyNode, ['outputs', 'capabilities'])
def check(token=None, **_):
    for item in token.node.value:
        if token.prev.node.value == 'outputs':
            output_obj = CfyOutput(item)
            if not output_obj.name and not output_obj.mapping:
                continue
        elif token.prev.node.value == 'capabilities':
            output_obj = CfyCapability(item)
        if output_obj.not_output():
            continue
        if isinstance(output_obj, CfyOutput):
            ctx['outputs'].update(output_obj.__dict__())
        else:
            ctx['capabilities'].update(output_obj.__dict__())
        desig = 'output' if token.prev.node.value == 'outputs' \
            else 'capability'
        if not output_obj.value:
            yield LintProblem(
                token.line, None,
                '{} {} does not provide a value.'.format(
                    desig, output_obj.name)
            )
        if not output_obj.description:
            yield LintProblem(
                token.line, None,
                '{} {} does not provide a description.'.format(
                    desig, output_obj.name)
            )


class CfyOutput(object):
    def __init__(self, nodes):
        self.name, self.mapping = self.get_name_mapping(nodes)
        if self.name and self.mapping:
            for key in list(self.mapping.keys()):
                if key not in ['description', 'value']:
                    del self.mapping[key]
            self.value = self.mapping.get('value')
            self.description = self.mapping.get('description')

    def get_name_mapping(self, nodes):
        return get_output(nodes)

    def not_output(self):
        return all([not k for k in self.mapping.values()])

    def __dict__(self):
        return {
            self.name: self.mapping
        }


class CfyCapability(CfyOutput):

    def get_name_mapping(self, nodes):
        return get_capability(nodes)

    def not_capability(self):
        return all([not k for k in self.mapping.values()])


def get_capability(nodes):
    if len(nodes) != 2:
        name = None
        mapping = None
    else:
        name = get_node_name(nodes[0])
        mapping = get_capability_mapping(nodes[1])
    return name, mapping


def get_output(nodes):
    if len(nodes) != 2:
        name = None
        mapping = None
    else:
        name = get_node_name(nodes[0])
        mapping = get_capability_mapping(nodes[1])
    return name, mapping


def get_node_name(node):
    if isinstance(node, yaml.nodes.ScalarNode):
        return node.value


def get_capability_mapping(node):
    mapping = {
        'description': None,
        'value': None,
    }
    if isinstance(node, yaml.nodes.MappingNode):
        for tup in node.value:
            if not len(tup) == 2:
                continue
            mapping_name = tup[0].value
            mapping_value = get_mapping_value(mapping_name, tup[1].value)
            mapping[mapping_name] = mapping_value
    return mapping


def get_mapping_value(name, value):
    if name not in ['value']:
        return value
    else:
        return recurse_mapping(value)



########
# Copyright (c) 2014-2022 Cloudify Platform Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import yamllint.rules
from ..utils import update_dict_values_recursive
from yamllint.config import YamlLintConfig
from yamllint.config import (
    validate_rule_conf,
    YamlLintConfigError)

from .constants import DEFAULT_CLOUDIFY_YAMLLINT_CONFIG


class YamlLintConfigExt(YamlLintConfig):
    def __init__(self, content=None, file=None, yamllint_rules=None):
        if content:
            update_dict_values_recursive(
                DEFAULT_CLOUDIFY_YAMLLINT_CONFIG, content)
        self._yamllint_rules = yamllint_rules or yamllint.rules
        super().__init__(DEFAULT_CLOUDIFY_YAMLLINT_CONFIG, file)

    @property
    def yamllint_rules(self):
        return self._yamllint_rules

    @yamllint_rules.setter
    def yamllint_rules(self, value):
        self._yamllint_rules = value

    def enabled_rules(self, filepath):
        return [self.yamllint_rules.get(id) for id, val in self.rules.items()
                if val is not False and (
                    filepath is None or 'ignore' not in val or
                    not val['ignore'].match_file(filepath))]

    def validate(self):
        for id in self.rules:
            try:
                rule = self.yamllint_rules.get(id)
            except Exception as e:
                raise YamlLintConfigError('invalid config: %s' % e)

            self.rules[id] = validate_rule_conf(rule, self.rules[id])



########
# Copyright (c) 2014-2022 Cloudify Platform Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import yaml
from .. import LintProblem

from . import constants
from ..generators import CfyNode
from .node_templates import recurse_node_template
from ..utils import process_relevant_tokens

VALUES = []

ID = 'relationships'
TYPE = 'token'
CONF = {'allowed-values': list(VALUES), 'check-keys': bool}
DEFAULT = {'allowed-values': ['true', 'false'], 'check-keys': True}
LIFECYCLE_OPS = {'preconfigure', 'postconfigure', 'establish', 'unlink'}
OP_KEYS = {'implementation', 'inputs'}


@process_relevant_tokens(CfyNode, 'relationships')
def check(token=None, **_):
    # raise Exception(token.node)
    relationship_type = CfyRelationshipType(token.node)
    if relationship_type.is_relationship_type:
        yield from check_relationship_types(relationship_type, token.line)
        return
    yield from relationships_not_list(token.node, token.line)
    for list_item in token.node.value:
        if isinstance(list_item, tuple) or isinstance(
                list_item.value, dict):
            yield from relationship_not_dict(list_item)
            continue
        is_target = False
        is_type = False
        for tup in list_item.value:
            if not len(tup) == 2:
                yield LintProblem(
                    list_item.value.start_mark.line + 1,
                    None,
                    "relationship dict must contain two entries, "
                    "type and target "
                    "The provided type is {}".format(type(len(tup)))
                )
            if tup[0].value == 'target':
                is_target = True
                yield from relationship_target_not_exist(
                    token, tup[1].value, tup[1].start_mark.line)
            elif tup[0].value == 'type':
                is_type = True
                yield from deprecated_type(
                    tup[1].value, tup[1].end_mark.line)
        yield from no_type(is_type, tup[1].start_mark.line)
        yield from no_target(is_target, tup[1].start_mark.line)


def no_type(type_bool, line):
    if not type_bool:
        yield LintProblem(
            line,
            None,
            "no relationship type provided. "
        )


def no_target(target, line):
    if not target:
        yield LintProblem(
            line,
            None,
            "no relationship target provided. "
        )


def deprecated_type(type_name, line):
    if type_name in constants.deprecated_relationship_types:
        yield LintProblem(
            line,
            None,
            "deprecated relationship type. "
            "Replace usage of {} with {}.".format(
                type_name,
                constants.deprecated_relationship_types[type_name]))


def relationships_not_list(node, line):
    if not isinstance(node, yaml.nodes.SequenceNode):
        yield LintProblem(
            line,
            None,
            "relationships block must be a list. "
            "The provided type is {}".format(type(node.value).mro()[0])
        )


def relationship_not_dict(list_item):
    if not isinstance(list_item, yaml.nodes.MappingNode):
        if isinstance(list_item, tuple):
            yield LintProblem(
                list_item[0].start_mark.line,
                None,
                "relationship must be a dict. "
                "The provided type is {}".format(type(list_item).mro()[0])
            )
        else:
            yield LintProblem(
                list_item.start_mark.line,
                None,
                "relationship must be a dict. "
                "The provided type is {}".format(type(list_item).mro()[0])
            )


def relationship_target_not_exist(token, target, line):
    if target not in token.node_templates:
        yield LintProblem(
            line,
            None,
            "relationship target node instance does not exist. "
            "The provided target is {}. Possible options are: {}.".format(
                target,
                [k for k in token.node_templates.keys()])
        )


class CfyRelationshipType(object):

    def __init__(self, node):
        self._node = node
        self._is_relationship_type = None
        self._name = None
        self.parsed = recurse_node_template(self._node)

        try:
            for k, v in self.parsed.items():
                self.name = k
                self.definition = v or {}
                break
        except (AttributeError, ValueError, KeyError):
            self.is_relationship_type = False
        else:
            self.is_relationship_type = True
            self.derived_from = self.definition.get('derived_from')
            self.connection_type = self.definition.get('connection_type')
            self.source_interfaces = self.definition.get('source_interfaces')
            self.target_interfaces = self.definition.get('target_interfaces')

    @property
    def interfaces(self):
        return [CfyRelationshipInterface(self.source_interfaces),
                CfyRelationshipInterface(self.target_interfaces)]


class CfyRelationshipInterfaces(object):
    def __init__(self, source, target):
        self.source = CfyRelationshipInterface(source)
        self.target = CfyRelationshipInterface(target)


class CfyRelationshipInterface(object):

    def __init__(self, data):
        self._data = data or {}

    @property
    def lifecycle(self):
        return self._data.get('cloudify.interfaces.relationship_lifecycle')


def check_relationship_types(relationship_type, line):
    if not relationship_type.name.startswith('cloudify.relationships.'):
        yield LintProblem(
            line,
            None,
            'relationship type name "{}" should '
            'start with "cloudify.relationships."'.format(
                relationship_type.name)
        )
    for interface in relationship_type.interfaces:
        if interface.lifecycle:
            keys = interface.lifecycle.keys()
            if not set(keys).issubset(LIFECYCLE_OPS):
                unexpected = set(keys) - LIFECYCLE_OPS
                yield LintProblem(
                    line,
                    None,
                    'unexpected {} in cloudify.interfaces.lifecycle: '
                    '{}'.format(
                        'operation' if len(unexpected) == 1 else 'operations',
                        unexpected
                    )
                )
            for op in LIFECYCLE_OPS:
                keys = interface.lifecycle.get(op, {}).keys()
                if not set(keys).issubset(OP_KEYS):
                    unexpected = set(keys) - OP_KEYS
                    yield LintProblem(
                        line,
                        None,
                        'unexpected key in {} operation definition: {}'.format(
                            op, unexpected
                        )
                    )
                elif op in interface.lifecycle.keys() and \
                        'implementation' not in keys:
                    yield LintProblem(
                        line,
                        None,
                        '{} operation definition does not declare '
                        'required implementation'.format(op)
                    )
